#include <memory>

#include "CommandBase.h"
#include "Robot.h"
#include "Subsystems/Drive.h"
#include "Commands/GrpAutoCenterLift.h"
#include "Commands/GrpAutoLeftLift.h"
#include "Commands/GrpAutoRightLift.h"
#include "Commands/GrpAutoCenter.h"
#include "Commands/GrpAutoLeft.h"
#include "Commands/GrpAutoRight.h"
#include "Commands/GrpAutoRedShoot.h"
#include "Commands/GrpAutoBlueShoot.h"

std::unique_ptr<RobotLog> Robot::robotLog = std::make_unique<RobotLog>();
std::unique_ptr<Dashboard> Robot::dashboard = std::make_unique<Dashboard>(13, 24, 1);
//double* Robot::dashValue;

void Robot::RobotInit() {
//	Robot::dashValue = new double[24];
//	for(int i = 0; i < 25; i++) {
//		Robot::dashValue[i] = 0;
//	}

	m_tunePID = new PIDControl();
	m_autoCommand = nullptr;
	m_autoDelay = 0;
	m_autoSelect = 0;
	m_cameraLED = new Solenoid(4);

	StartCamera();

	CommandBase::drive->InitDrive();
	CommandBase::fuelHandler->InitFuelHandler();
	CommandBase::gearPickup->InitGearPickup();
	CommandBase::goat->InitGoat();
	CommandBase::vision->InitVision();

	CommandBase::drive->CalibrateGyro();
}

void Robot::DisabledInit() {
	robotLog->SetMode(RobotLog::mDisabled);
	robotLog->Write("");
	sprintf(m_log, "Miguelito: Disabled Init %s", dashboard->GetTimeStamp().c_str());
	robotLog->Write(m_log);

	dashboard->SetRobotMode(1);

	CommandBase::drive->ResetNavigation();
	CommandBase::drive->SetDriveMode(Drive::dmDriveRotate);
	CommandBase::fuelHandler->SetAgitator(FuelHandler::rsOff);
	CommandBase::fuelHandler->SetShooter(false);
	CommandBase::fuelHandler->SetLegs(FuelHandler::flLeft, false);
	CommandBase::fuelHandler->SetLegs(FuelHandler::flRight, false);
	CommandBase::gearPickup->SetGearPosition(false);
	CommandBase::gearPickup->SetRollerMode(GearPickup::rmOff);
	CommandBase::gearPickup->SetAirVent(false);

//	m_aimCamera.SetExposureManual(10);
//	m_cameraLED->Set(true);

	robotLog->Write("Miguelito: Disabled Periodic");
	robotLog->Close();
}

void Robot::DisabledPeriodic() {
//	static int cameraCounter = 250;
//
//	if (cameraCounter > 0) {
//		cameraCounter--;
//		if (cameraCounter <= 0) {
//			CommandBase::vision->StartCamera("10.5.25.9");
//			robotLog->Write("Camera started");
//		}
//	}

	frc::Scheduler::GetInstance()->Run();
	if(dashboard->GetDashButtonPress(0, dbGyroCalibrate)) {
		CommandBase::drive->CalibrateGyro();
	}

	CommandBase::drive->SetDashboardValues();
	CommandBase::fuelHandler->SetDashboardValues();
	CommandBase::gearPickup->SetDashboardValues();
	CommandBase::goat->SetDashboardValues();
}


void Robot::AutonomousInit() {
	robotLog->SetMode(RobotLog::mAutonomous);
	robotLog->Write("");
	sprintf(m_log, "Miguelito: Autonomous Init %s", dashboard->GetTimeStamp().c_str());
	robotLog->Write(m_log);

	dashboard->SetRobotMode(2);
	CommandBase::fuelHandler->SetAgitator(FuelHandler::rsForward);

	m_autoSelect = (int)dashboard->GetDashValue(dvAutoSelect);
	m_autoDelay = dashboard->GetDashValue(dvAutoDelay) / 4.0;

	if (m_autoCommand != nullptr) delete m_autoCommand;

	switch(m_autoSelect) {
		case 1:  m_autoCommand = new GrpAutoLeftLift(m_autoDelay);
				 break;
		case 2:  m_autoCommand = new GrpAutoCenterLift(m_autoDelay);
				 break;
		case 3:  m_autoCommand = new GrpAutoRightLift(m_autoDelay);
				 break;
		case 4:  m_autoCommand = new GrpAutoLeft(m_autoDelay);
				 break;
		case 5:  m_autoCommand = new GrpAutoCenter(m_autoDelay);
				 break;
		case 6:  m_autoCommand = new GrpAutoRight(m_autoDelay);
				 break;
		case 7:	 m_autoCommand = new GrpAutoRedShoot(m_autoDelay);
				 break;
		case 8:	 m_autoCommand = new GrpAutoBlueShoot(m_autoDelay);
				 break;
		default: m_autoCommand = nullptr;
	}

	if (m_autoCommand != nullptr) m_autoCommand->Start();

	sprintf(m_log, "Miguelito: Autonomous Periodic Selection=%d  Delay=%f", m_autoSelect, m_autoDelay);
	robotLog->Write(m_log);
}

void Robot::AutonomousPeriodic() {
	robotLog->StartPeriodic();

	frc::Scheduler::GetInstance()->Run();

	CommandBase::drive->Periodic();
	CommandBase::fuelHandler->Periodic();
	CommandBase::gearPickup->Periodic();
	CommandBase::goat->Periodic();

	robotLog->EndPeriodic();
}

void Robot::TeleopInit() {
	robotLog->SetMode(RobotLog::mTeleop);
	robotLog->Write("");
	sprintf(m_log, "Miguelito: Teleop Init %s", dashboard->GetTimeStamp().c_str());
	robotLog->Write(m_log);

	if (m_autoCommand != nullptr) m_autoCommand->Cancel();
//	m_aimCamera.SetExposureAuto();

	dashboard->SetRobotMode(3);
	CommandBase::fuelHandler->SetAgitator(FuelHandler::rsForward);

	robotLog->Write("Miguelito: Teleop Periodic");
}

void Robot::TeleopPeriodic() {
	robotLog->StartPeriodic();

	frc::Scheduler::GetInstance()->Run();

	CommandBase::drive->Periodic();
	CommandBase::fuelHandler->Periodic();
	CommandBase::gearPickup->Periodic();
	CommandBase::goat->Periodic();

	robotLog->EndPeriodic();
}

void Robot::TestInit() {
	robotLog->SetMode(RobotLog::mTest);
	robotLog->Write("");
	sprintf(m_log, "Miguelito: Test Init %s", dashboard->GetTimeStamp().c_str());
	robotLog->Write(m_log);

	CommandBase::gearPickup->SetAirVent(true);

	dashboard->SetRobotMode(4);
	robotLog->Write("Miguelito: Test Periodic");
}

void Robot::TestPeriodic() {
	static double   setpoint = 0;
	static int 		tunePID  = -1;

	if(dashboard->GetDashButton(0, dbRunPid)) {
		if(tunePID < 0 ) {
			tunePID  = dashboard->GetDashValue(dvPidSelect);
			setpoint = dashboard->GetDashValue(dvPidSetpoint);

			switch (tunePID) {
				case 0: setpoint += CommandBase::drive->GetDriveDistance();
						CommandBase::drive->InitRotatePID();
						break;
				case 1: setpoint += CommandBase::drive->GetGyroAngle();
						break;
				default:;
			}

			m_tunePID->SetSetpoint(0);
			m_tunePID->SetOutputRange(-dashboard->GetDashValue(dvPidMaxOut),
									   dashboard->GetDashValue(dvPidMaxOut));

			m_tunePID->SetCoefficient('P', dashboard->GetDashValue(dvPthreshold),
										   dashboard->GetDashValue(dvPabove),
										   dashboard->GetDashValue(dvPbelow));
			m_tunePID->SetCoefficient('I', dashboard->GetDashValue(dvIthreshold),
										   dashboard->GetDashValue(dvIabove),
										   dashboard->GetDashValue(dvIbelow));
			m_tunePID->SetCoefficient('D', dashboard->GetDashValue(dvDthreshold),
										   dashboard->GetDashValue(dvDabove),
										   dashboard->GetDashValue(dvDbelow));
			m_tunePID->Reset();
		}

		switch(tunePID)  {
			case 0: CommandBase::drive->TuneDrivePID(m_tunePID->Calculate(CommandBase::drive->ErrorDrive(setpoint), true));
					break;
			case 1: CommandBase::drive->TuneRotatePID(m_tunePID->Calculate(CommandBase::drive->ErrorRotate(setpoint), true));
					break;
			default:;
		}

		CommandBase::fuelHandler->Periodic();
		CommandBase::gearPickup->Periodic();

	} else {
		if(tunePID >= 0) tunePID = -1;

		CommandBase::drive->Periodic();
		CommandBase::fuelHandler->Periodic();
		CommandBase::gearPickup->Periodic();
		CommandBase::goat->Periodic(true);
	}
}

void Robot::StartCamera() {

	CameraServer* cameraServer = CameraServer::GetInstance();

	cs::UsbCamera	usbCamera = cameraServer->StartAutomaticCapture(1);
	usbCamera.SetResolution(320, 240);
	usbCamera.SetFPS(20);

//	m_aimCamera = cameraServer->StartAutomaticCapture("aimCamera", 0);
//	m_aimCamera.SetResolution(320, 240);
//	m_aimCamera.SetFPS(20);
//	m_aimCamera.SetBrightness(0);
//	m_aimCamera.SetExposureManual(10);

	cs::UsbCamera	usbCamera1 = cameraServer->StartAutomaticCapture("aimCamera", 0);
	usbCamera1.SetResolution(320, 240);
	usbCamera1.SetFPS(20);
	usbCamera1.SetBrightness(0);
	usbCamera1.SetExposureManual(10);

//	cs::AxisCamera	axisCamera = cameraServer->AddAxisCamera("10.5.25.9");
}

START_ROBOT_CLASS(Robot)







