#include "Drive.h"
#include "../Robot.h"
#include "../CommandBase.h"

Drive::Drive() : Subsystem("Drive") {
	m_motorLF 		= nullptr;
	m_motorLR 		= nullptr;
	m_motorRF 		= nullptr;
	m_motorRR 		= nullptr;

	m_gyro			= nullptr;

	m_driveEncoderL	= nullptr;
	m_driveEncoderR = nullptr;

	m_drivePID		= nullptr;
	m_rotatePID		= nullptr;

	m_autoState 	= asNull;
	m_driveMode		= dmDriveRotate;
	m_driveWheels 	= dwAll;
	m_targetAngle	= 0;
	m_targetDrive	= 0;
	m_timeoutCounter= 0;
	m_useVision		= false;
}

void Drive::CalibrateGyro() {
	Robot::dashboard->SetRobotValue(Robot::rvGyroCal, 1);
	m_gyro->Calibrate();
	m_gyro->Reset();
	Robot::dashboard->SetRobotValue(Robot::rvGyroCal, 0);
}

std::string Drive::DriveModeName(DriveMode mode) {
	switch (mode) {
		case dmDriveRotate:		return "Drive-Rotate";
		case dmDriveStrafe:		return "Drive-Strafe";
		case dmStrafe:			return "Strafe";
		case dmJoystick:		return "Joystick";
		case dmAutoRotate:		return "Auto Rotate";
		case dmAutoDrive:		return "Auto Drive";
		case dmAutoStrafe:		return "Auto Strafe";
		case dmAutoDone:		return "Auto Completed";
		case dmGearPlace:		return "Gear Place";
		default:	return "?";
	}
}

double Drive::ErrorDrive(double target) {
	return(target - m_driveEncoderL->GetDistance());
}

double Drive::ErrorRotate(double target) {
	double error = GetGyroAngle() - target;

	if (error > 180) {
		error -= 360;
	} else if (error < -180) {
		error += 360;
	}
	//error = (error > 180) ? error - 360 : (error < -180) ? error + 360 : error;

	return error;
}

Drive::AutoState Drive::GetAutoState(){
	return m_autoState;
}

double Drive::GetDriveDistance() {
	return m_driveEncoderL->GetDistance();
}

double Drive::GetDriveRate() {
	return m_driveEncoderL->GetRate();
}

double Drive::GetGyroAngle() {
	double angle = ((int) (m_gyro->GetAngle() * 10) % 3600) / 10.0;
	if (angle < 0) angle += 360;
	return angle;
}

void Drive::InitAutoDrive(double distance, bool useAbsolute, double maxDrive, int msecTimeout, bool useVision){
	m_useVision = useVision;

	if(m_useVision) {
		double visionAngle = CommandBase::vision->GetTargetAngle();
		double visionDistance = -CommandBase::vision->GetTargetDistance();

		if(visionDistance != 0) m_targetDrive = m_driveEncoderL->GetDistance() + visionDistance;
		else					m_targetDrive = m_driveEncoderL->GetDistance() + distance;

		InitRotatePID(visionAngle);

	} else {
		if(useAbsolute) m_targetDrive = distance;
		else			m_targetDrive = m_driveEncoderL->GetDistance() + distance;
		InitRotatePID();
	}

	m_drivePID->SetOutputRange(-maxDrive, maxDrive);

	if(msecTimeout <= 0) m_timeoutCounter = -1;
	else				 m_timeoutCounter = msecTimeout / 20;

	m_autoState = asInProgress;
	m_driveMode = dmAutoDrive;

	sprintf(m_log, "Drive:   Auto Drive Init: Start=%6.1f End=%6.1f Heading=%5.1f",
			m_driveEncoderL->GetDistance(), m_targetDrive, m_targetAngle);
	Robot::robotLog->Write(m_log);
	Robot::dashboard->SetDashValue(Robot::rvDriveMode, dmAutoDrive);
}

void Drive::InitAutoRotate(double angle, bool useAbsolute, double maxRotate, int msecTimeout, DriveWheels wheels){
	if(useAbsolute) m_targetAngle = angle;
	else			m_targetAngle = GetGyroAngle() + angle;

	m_rotatePID->SetOutputRange(-maxRotate, maxRotate);

	if(msecTimeout <= 0) m_timeoutCounter = -1;
	else 				 m_timeoutCounter = msecTimeout / 20;

	m_autoState = asInProgress;
	m_driveMode = dmAutoRotate;
	m_driveWheels = wheels;

	sprintf(m_log, "Drive:   Auto Rotate Init: Angle=%5.1f", m_targetAngle);
	Robot::robotLog->Write(m_log);
	Robot::dashboard->SetDashValue(Robot::rvDriveMode, dmAutoRotate);
}

void Drive::InitAutoStrafe(double distance, bool useAbsolute, double maxStrafe, int msecTimeout){
	if(useAbsolute) m_targetDrive = distance;
	else 			m_targetDrive = m_driveEncoderL->GetDistance() + distance;

	m_drivePID->SetOutputRange(-maxStrafe, maxStrafe);

	if(msecTimeout <= 0) m_timeoutCounter = -1;
	else				 m_timeoutCounter = msecTimeout / 20;

	InitRotatePID();
	m_autoState = asInProgress;
	m_driveMode = dmAutoStrafe;

	sprintf(m_log, "Drive:   Init Auto Strafe: Distance=%6.1f Heading=%5.1f", m_targetDrive, m_targetAngle);
	Robot::robotLog->Write(m_log);
	Robot::dashboard->SetDashValue(Robot::rvDriveMode, dmAutoStrafe);
}

void Drive::InitDefaultCommand() {
}

void Drive::InitDrive() {
	m_motorLF = new VictorSP(Robot::pwmDriveMotorLF);
	m_motorLR = new VictorSP(Robot::pwmDriveMotorLR);
	m_motorRF = new VictorSP(Robot::pwmDriveMotorRF);
	m_motorRR = new VictorSP(Robot::pwmDriveMotorRR);

	m_gyro	  = new ADXRS450_Gyro();

	m_driveEncoderL = new Encoder(Robot::dioDriveEncoderLA, Robot::dioDriveEncoderLB, false, Encoder::k4X);
	m_driveEncoderL->SetDistancePerPulse(0.006085);
	m_driveEncoderL->Reset();

	m_driveEncoderR = new Encoder(Robot::dioDriveEncoderRA, Robot::dioDriveEncoderRB, true, Encoder::k4X);
	m_driveEncoderR->SetDistancePerPulse(0.04868);
	m_driveEncoderR->Reset();

	m_drivePID = new PIDControl();
	m_drivePID->SetCoefficient('P', 0, 0.02, 0);
	m_drivePID->SetCoefficient('I', 12, 0, 0.001);
	m_drivePID->SetCoefficient('D', 24, 0, 0.35);
	m_drivePID->SetInputRange(-180.0, 180.0);
	m_drivePID->SetOutputRange(-0.7, 0.7);
	m_drivePID->SetSetpoint(0);

	m_rotatePID = new PIDControl();
	m_rotatePID->SetCoefficient('P', 0, 0.04, 0);
	m_rotatePID->SetCoefficient('I', 10, 0, 0.005);
	m_rotatePID->SetCoefficient('D', 0, 0.1, 0);
	m_rotatePID->SetInputRange(-180.0, 180.0);
	m_rotatePID->SetOutputRange(-0.7, 0.7);
	m_rotatePID->SetSetpoint(0);

	SetDriveMode(dmDriveRotate);
}

void Drive::InitRotatePID(double addAngle) {
	m_targetAngle = GetGyroAngle() + addAngle;
	m_rotatePID->Reset();
	m_rotatePID->SetOutputRange(-0.7, 0.7);
}

double Drive::Limit(double Value) {
	if(Value > 1.0) return 1.0;
	if(Value < -1.0) return -1.0;
	return Value;
}

void Drive::MecanumDrive(double drive, double strafe, double rotate, DriveWheels driveWheels) {
	double motorAbs = 0;
	double motorPwm[4];
	double motorMax = 0;

	motorPwm[0] = rotate + drive + strafe;		// Left Front
	motorPwm[1] = rotate + drive - strafe;		// Left Rear
	motorPwm[2] = rotate - drive + strafe;		// Right Front
	motorPwm[3] = rotate - drive - strafe;		// Right Rear

	for(int i = 0; i < 4; i++) {
		motorAbs = fabs(motorPwm[i]);
		if(motorMax < motorAbs) motorMax = motorAbs;
	}

	if(motorMax > 1.0) {
		for(int i = 0; i < 4; i++){
			motorPwm[i] = Limit(motorPwm[i] / motorMax);
		}
	}

	switch(driveWheels) {
		case dwAll:
			m_motorLF->Set(motorPwm[0]);
			m_motorLR->Set(motorPwm[1]);
			m_motorRF->Set(motorPwm[2]);
			m_motorRR->Set(motorPwm[3]);
			break;
		case dwRear:
			m_motorLF->Set(0);
			m_motorLR->Set(motorPwm[1]);
			m_motorRF->Set(0);
			m_motorRR->Set(motorPwm[3]);
			break;
		case dwFront:
			m_motorLF->Set(motorPwm[0]);
			m_motorLR->Set(0);
			m_motorRF->Set(motorPwm[2]);
			m_motorRR->Set(0);
			break;
		case dwPivotLeft:
			m_motorLF->Set(0);
			m_motorLR->Set(motorPwm[1]);
			m_motorRF->Set(motorPwm[2]);
			m_motorRR->Set(motorPwm[3]);
			break;
		case dwPivotRight:
			m_motorLF->Set(motorPwm[0]);
			m_motorLR->Set(motorPwm[1]);
			m_motorRF->Set(0);
			m_motorRR->Set(motorPwm[3]);
			break;
		default:;
	}
}

void Drive::Periodic() {
	double errorDrive 	= 0;
	double errorRotate  = 0;
	bool   isFinished   = false;
	double pwmDrive    	= 0;
	double pwmRotate   	= 0;
	double pwmStrafe   	= 0;

	if (m_driveMode != dmAutoRotate) m_driveWheels = dwAll;

	switch (m_driveMode){
		case dmDriveRotate:
			pwmDrive = -CommandBase::oi->GetDriveYAxis();
			pwmRotate = -CommandBase::oi->GetDriveXAxis();
			break;

		case dmDriveStrafe:
			pwmDrive = -CommandBase::oi->GetDriveYAxis();
			pwmStrafe = CommandBase::oi->GetDriveXAxis();
			pwmRotate = -m_rotatePID->Calculate(ErrorRotate(m_targetAngle), false);
			break;

		case dmStrafe:
			pwmStrafe = CommandBase::oi->GetDriveXAxis();
			pwmRotate = -m_rotatePID->Calculate(ErrorRotate(m_targetAngle),false);
			break;

		case dmJoystick:
			pwmDrive = CommandBase::oi->GetDriveYAxis();
			pwmRotate = CommandBase::oi->GetDriveZAxis();
			pwmStrafe = CommandBase::oi->GetDriveXAxis();
			break;

		case dmAutoRotate:
			errorRotate = ErrorRotate(m_targetAngle);

			if(fabs(errorRotate) <= 1) {
				if (m_autoState != asFinished) {
					m_autoState = asFinished;
					sprintf(m_log, "Drive:   Auto Rotate Done: Angle=%5.1f", GetGyroAngle());
					Robot::robotLog->Write(m_log);
				}
			} else if(m_timeoutCounter == -1) {
				pwmRotate = -m_rotatePID->Calculate(errorRotate, false);
			} else if(m_timeoutCounter > 0) {
				pwmRotate = -m_rotatePID->Calculate(errorRotate, false);
				m_timeoutCounter--;
			} else {
				m_autoState = asTimedOut;
				Robot::robotLog->Write("Drive:   Auto Rotate Timeout");
			}
			break;

		case dmAutoDrive:
			errorDrive = ErrorDrive(m_targetDrive);
			isFinished = (fabs(m_driveEncoderL->GetDistance()) >= fabs(m_targetDrive));

			if (isFinished && m_useVision) {
				isFinished  = (fabs(m_driveEncoderL->GetRate()) < 1.0);
			}

			if (isFinished) {
				if (m_autoState != asFinished) {
					m_autoState = asFinished;
					sprintf(m_log, "Drive:   Auto Drive Done: Distance=%6.1f Heading=%5.1f",
							m_driveEncoderL->GetDistance(), GetGyroAngle());
					Robot::robotLog->Write(m_log);
				}

			} else if(m_timeoutCounter == -1) {
				pwmDrive = m_drivePID->Calculate(errorDrive, false);

			} else if(m_timeoutCounter > 0) {
				pwmDrive = m_drivePID->Calculate(errorDrive, false);
				if (m_useVision && pwmDrive < 0.1) pwmDrive = 0.1;
				m_timeoutCounter--;

//				if (m_useVision) {
//					if (pwmDrive < 0) {
//						if (pwmDrive > -0.2) pwmDrive = -0.2;
//					} else if (pwmDrive < 0.2) {
//						pwmDrive = 0.2;
//					}
//				}
			} else {
				m_autoState = asTimedOut;
				sprintf(m_log, "Drive:   Auto Drive Timeout: Distance=%6.1f Heading=%5.1f",
						m_driveEncoderL->GetDistance(), GetGyroAngle());
				Robot::robotLog->Write(m_log);
			}

			pwmRotate = -m_rotatePID->Calculate(ErrorRotate(m_targetAngle), false);
			break;

		case dmAutoStrafe:
			errorDrive = ErrorDrive(m_targetDrive);

			if(fabs(errorDrive) <= 2) {
				m_autoState = asFinished;
				Robot::robotLog->Write("Drive:   Auto Strafe Completed");
			} else if(m_timeoutCounter == -1) {
				pwmStrafe = m_drivePID->Calculate(errorDrive, false);
			} else if(m_timeoutCounter > 0) {
				pwmStrafe = m_drivePID->Calculate(errorDrive, false);
				m_timeoutCounter--;
			} else {
				m_autoState = asTimedOut;
				Robot::robotLog->Write("Drive:   Auto Strafe Timeout");
			}

			pwmRotate = -m_rotatePID->Calculate(ErrorRotate(m_targetAngle), false);
			break;

		case dmGearPlace:
			if(m_timeoutCounter > 0){
				pwmDrive = -0.25;					// -0.2
				m_timeoutCounter--;
			} else {
				pwmDrive = 0;
				m_autoState = asTimedOut;
				SetDriveMode(dmAutoDone);
			}
			break;
		default:;
	}

	MecanumDrive(pwmDrive, pwmStrafe, pwmRotate, m_driveWheels);

	SetDashboardValues();
}

void Drive::ResetNavigation() {
	m_gyro->Reset();
	m_driveEncoderL->Reset();
	m_driveEncoderR->Reset();
}

void Drive::SetDashboardValues() {
	Robot::dashboard->SetRobotValue(Robot::rvDriveGyro, GetGyroAngle());
	Robot::dashboard->SetRobotValue(Robot::rvDriveEncoderL, m_driveEncoderL->GetDistance() * 10);
	Robot::dashboard->SetRobotValue(Robot::rvDriveEncoderR, m_driveEncoderR->GetDistance() * 10);
}

void Drive::SetDriveMode(DriveMode mode, int gearPlaceTimeout) {
	if (m_driveMode != mode) {
		m_driveMode = mode;

		if (mode == dmStrafe || mode == dmDriveStrafe) {
			InitRotatePID();
		} else if (mode == dmAutoDone) {
			m_driveMode = dmDriveRotate;
		} else if (mode == dmGearPlace) {
			m_timeoutCounter = gearPlaceTimeout;
			m_autoState = asInProgress;
		}

		sprintf(m_log, "Drive:   Drive Mode=%s", DriveModeName(mode).c_str());
		Robot::robotLog->Write(m_log);
		Robot::dashboard->SetRobotValue(Robot::rvDriveMode, (double)m_driveMode);
	}
}

void Drive::TuneDrivePID(double pidOutput) {
	MecanumDrive(pidOutput, 0, -m_rotatePID->Calculate(ErrorRotate(m_targetAngle),false), dwAll);
}

void Drive::TuneRotatePID(double pidOutput) {
	MecanumDrive(0, 0, -pidOutput, dwAll);
}

void Drive::TuneStrafePID(double pidOutput) {
	MecanumDrive(0, pidOutput, -m_rotatePID->Calculate(ErrorRotate(m_targetAngle), false), dwAll);
}
