#include "FuelHandler.h"
#include "../Robot.h"
#include "CommandBase.h"

FuelHandler::FuelHandler() : Subsystem("Fuel Handler") {
	m_fuelAgitator  = nullptr;
	m_fuelAuger		= nullptr;
	m_fuelChurner	= nullptr;
	m_fuelIntake 	= nullptr;
	m_highShooter 	= nullptr;
	m_lowShooter  	= nullptr;
	m_panel			= nullptr;
	m_leftLeg		= nullptr;
	m_rightLeg		= nullptr;

	m_agitatorState = rsOff;
	m_augerState	= rsOff;
	m_churnerState	= rsOff;
	m_intakeMode 	= imOff;
	m_leftLegDown 	= false;
	m_rightLegDown 	= false;
	m_shooterDelay  = 0;
	m_shooterOn		= false;
}

void FuelHandler::InitDefaultCommand() {
}

void FuelHandler::InitFuelHandler() {
	m_fuelAgitator = new Relay(Robot::relFuelAgitator, Relay::kBothDirections);
	m_fuelAuger    = new Relay(Robot::relFuelAuger, Relay::kBothDirections);
	m_fuelChurner  = new VictorSP(Robot::pwmFuelChurner);
	m_fuelIntake   = new VictorSP(Robot::pwmFuelIntake);

	m_highShooter = new CANTalon(Robot::canHighShooter);
	m_highShooter->ConfigNeutralMode(CANTalon::NeutralMode::kNeutralMode_Coast);
	m_highShooter->SetFeedbackDevice(CANTalon::FeedbackDevice::CtreMagEncoder_Relative);
	m_highShooter->SetControlMode(CANTalon::ControlMode::kPercentVbus);
	m_highShooter->SetClosedLoopOutputDirection(true);
	m_highShooter->SetVoltageRampRate(48.0);
	m_highShooter->SelectProfileSlot(0);
	m_highShooter->SetP(0.05);
	m_highShooter->SetI(0.0005);
	m_highShooter->SetD(0);
	m_highShooter->SetF(0.026);
	m_highShooter->SetIzone(3000);		// 2000

	m_lowShooter = new CANTalon(Robot::canLowShooter);
	m_lowShooter->ConfigNeutralMode(CANTalon::NeutralMode::kNeutralMode_Coast);
	m_lowShooter->SetFeedbackDevice(CANTalon::FeedbackDevice::CtreMagEncoder_Relative);
	m_lowShooter->SetControlMode(CANTalon::ControlMode::kPercentVbus);
	m_lowShooter->SetClosedLoopOutputDirection(true);
	m_lowShooter->SetVoltageRampRate(48.0);
	m_lowShooter->SelectProfileSlot(0);
	m_lowShooter->SetP(0.05);
	m_lowShooter->SetI(0.0005);
	m_lowShooter->SetD(0);
	m_lowShooter->SetF(0.026);
	m_lowShooter->SetIzone(3000);		// 2000

	m_panel		= new PowerDistributionPanel(0);
	m_leftLeg	= new Solenoid(Robot::solLeftLeg);
	m_rightLeg	= new Solenoid(Robot::solRightLeg);

	Robot::dashboard->SetRobotValue(Robot::rvFuelLegs, 0);
}

void FuelHandler::Periodic() {
	double intakePwm = 0;
	double intakeMax;
	double intakeMin;
	double rateMax;
	double rateNow;

	if(m_shooterDelay > 0){
		m_shooterDelay--;

		if(m_shooterDelay <= 0){
			m_lowShooter->Set(Robot::dashboard->GetDashValue(Robot::dvLowShooterSpeed));
			SetAuger(rsForward);
			SetChurner(rsOn);

			sprintf(m_log, "Fuel:    Low Shooter=%f", Robot::dashboard->GetDashValue(Robot::dvLowShooterSpeed));
			Robot::robotLog->Write(m_log);
		}
	}

	switch(m_intakeMode){
		case imIn:
//			intakeMax = Robot::dashValue[Robot::dvMaxFuelIntake];
			intakeMax = Robot::dashboard->GetDashValue(Robot::dvMaxFuelIntake);
			intakeMin = Robot::dashboard->GetDashValue(Robot::dvMinFuelIntake);
			rateMax  = Robot::dashboard->GetDashValue(Robot::dvMaxDriveSpeed);
			rateNow  = CommandBase::drive->GetDriveRate();

			if (rateNow > 0)	intakePwm = intakeMin + (rateNow / rateMax) * (intakeMax - intakeMin);
			else				intakePwm = intakeMin;
			break;

		case imOut:
			intakePwm = -0.5;
			break;

		default:;
	}

	m_fuelIntake->Set(-intakePwm);

	if (m_churnerState == rsOff)	m_fuelChurner->Set(0);
	else							m_fuelChurner->Set(Robot::dashboard->GetDashValue(Robot::dvFuelChurnerSpeed));

	SetDashboardValues();
}

void FuelHandler::SetAgitator(RelayState state) {
	switch(state) {
		case rsOff:
			m_fuelAgitator->Set(Relay::kOff);
			break;
		case rsForward:
			m_fuelAgitator->Set(Relay::kForward);
			break;
		case rsReverse:
			m_fuelAgitator->Set(Relay::kReverse);
			break;
		default:;
	}

	m_agitatorState = state;
}

void FuelHandler::SetAuger(RelayState state) {
	switch(state) {
		case rsOff:
			m_fuelAuger->Set((Relay::kOff));
			break;
		case rsForward:
			m_fuelAuger->Set(Relay::kForward);
			break;
		case rsReverse:
			m_fuelAuger->Set(Relay::kReverse);
			break;
		default:;
	}

	m_augerState = state;
}

void FuelHandler::SetChurner(RelayState state) {
	m_churnerState = state;
}

void FuelHandler::SetDashboardValues() {
	std::string data;

	Robot::dashboard->SetRobotValue(Robot::rvLowShooter, (double)m_lowShooter->GetEncVel() * 0.14648);
	Robot::dashboard->SetRobotValue(Robot::rvHighShooter, -(double)m_highShooter->GetEncVel() * 0.14648);
	Robot::dashboard->SetRobotValue(Robot::rvFuelIntake, m_fuelIntake->Get());

	if(Robot::dashboard->GetDashButton(0, Robot::dbLogData)) {
		data = Robot::dashboard->DataString(m_lowShooter->GetEncVel(), 2);
		data += Robot::dashboard->DataString(m_highShooter->GetEncVel(), 0);
		Robot::robotLog->WriteData(data);
	}
}
void FuelHandler::SetIntakeMode(IntakeMode mode) {
	m_intakeMode = mode;
}

void FuelHandler::SetLegs(FuelLegs legs, bool legDown) {
	int legState = 0;

	if (legs == flRight){
		if (m_rightLegDown != legDown) {
			m_rightLeg->Set(legDown);
			m_rightLegDown = legDown;

			if (m_rightLegDown) Robot::robotLog->Write("Fuel:    Right Legs=Down");
			else				Robot::robotLog->Write("Fuel:    Right Legs=Up");
		}
	} else {
		if (m_leftLegDown != legDown) {
			m_leftLeg->Set(legDown);
			m_leftLegDown = legDown;

			if (m_leftLegDown) 	Robot::robotLog->Write("Fuel:    Left Legs=Down");
			else				Robot::robotLog->Write("Fuel:    Left Legs=Up");
		}
	}

	if (m_leftLegDown) 	legState |= 1;
	if (m_rightLegDown)	legState |= 2;
	Robot::dashboard->SetRobotValue(Robot::rvFuelLegs, (double)legState);
}

void FuelHandler::SetShooter(bool shooterOn){
	 if(m_shooterOn != shooterOn){
		 if(shooterOn){
			 m_lowShooter->SetControlMode(CANTalon::ControlMode::kSpeed);
			 m_highShooter->SetControlMode(CANTalon::ControlMode::kSpeed);
			 m_lowShooter->Set(-1000);
			 m_highShooter->Set(-Robot::dashboard->GetDashValue(Robot::dvHighShooterSpeed));
			 m_shooterDelay = 25;

			 sprintf(m_log, "Fuel:    High Shooter=%f", Robot::dashboard->GetDashValue(Robot::dvHighShooterSpeed));
			 Robot::robotLog->Write(m_log);

		 } else {
			 m_lowShooter->SetControlMode(CANTalon::ControlMode::kPercentVbus);
			 m_highShooter->SetControlMode(CANTalon::ControlMode::kPercentVbus);
			 SetAuger(rsOff);
			 SetChurner(rsOff);

			 Robot::robotLog->Write("Fuel:    Shooter=Off");
		 }
	 }

	 m_shooterOn = shooterOn;
}
