#include "Vision.h"
#include "../Robot.h"
#include <CameraServer.h>

// Target Angle:	Tan = x / y where x is the normalized location of the target center (-1 to 1)
//					When:  x = 1 when the Target Angle is 1/2 of camera view angle  	(Axis=48.2/2; LifeCam=50.7/2)
//						   y = 1 / tan(25.35) = 2.11 					 				(Axis=2.2355; LifeCam=2.11)
//
//					Target Angle = atan(x, y)
//								 = atan(x, 2.11)
//
// Target Distance:	Image Inch width = Image pixel width * Target inch width / Target pixel width
//							 		 = 10.25 * 320 / Target pixel width
//							   		 = 3280 / Target pixel width
//					Distance = (Image inch width / 2) / Tan(Camera image angle / 2)
//							 = (1640 / Tan(25.35)) / Target pixel width
//							 = 3462 / Target pixel width

Vision::Vision() : Subsystem("Vision") {
	m_angleBias = 0;
	m_task = nullptr;
	m_searchState = ssDone;
	m_targetAngle = 0;
	m_targetDistance = 0;
}

void Vision::FindTarget(double angleBias) {
	Robot::robotLog->Write(("Vision:  Initiate Target search"));
	m_angleBias = angleBias;
	m_targetAngle = 0;
	m_targetDistance = 0;
	m_searchState = ssLooking;

	m_task = new std::thread(TargetSearch, this);
	m_task->detach();
}

bool Vision::GetCameraImage(cv::Mat& image) {
	if (m_sink.GrabFrame(image) == 0) return false;
	return true;
}

Vision::SearchState Vision::GetSearchState()
{
	return m_searchState;
}

double Vision::GetTargetAngle() {
	return m_targetAngle;
}

double Vision::GetTargetDistance() {
	return m_targetDistance;
}

void Vision::InitDefaultCommand() {
}

void Vision::InitVision() {
	m_sink = CameraServer::GetInstance()->GetVideo("aimCamera");
	m_sink.SetEnabled(true);
}

void Vision::SearchResults(bool targetFound, double targetAngle, double targetDistance) {
	if (targetFound) {
		m_searchState = ssTargetFound;
		m_targetAngle = targetAngle + m_angleBias;
		m_targetDistance = targetDistance;

		sprintf(m_log, "Vision:  Target Angle=%5.1f  Distance=%5.0f", m_targetAngle, m_targetDistance);
		Robot::robotLog->Write(m_log);
	} else {
		m_searchState = ssNoTarget;
		Robot::robotLog->Write("Vision:  No Target found");
	}
}

void Vision::TargetSearch(Vision* host) {
	cv::Mat	imageA;
	cv::Mat imageB;
	double  ratio;
	char	robotLog[100];
	double	targetAngle = 0;
	int		targetCount = 0;
	double  targetDistance = 0;

	if (host->GetCameraImage(imageA)) {
		cv::cvtColor(imageA, imageB, cv::COLOR_BGR2HSV);
		cv::inRange(imageB, cv::Scalar(0, 180, 130), cv::Scalar(180, 255, 255), imageA);

		std::vector<std::vector<cv::Point>> contours;
		std::vector<cv::Vec4i> hierarchy;
		contours.clear();

		cv::findContours(imageA, contours, hierarchy, cv::RETR_LIST, cv::CHAIN_APPROX_SIMPLE);

		std::vector<std::vector<cv::Point>> targets;
		targets.clear();

		for (std::vector<cv::Point> contour: contours) {
			cv::Rect rect = boundingRect(contour);
			double area = cv::contourArea(contour);
			if (area < 125) continue;

			ratio = (double) rect.width / (double) rect.height;

			sprintf(robotLog, "Vision:  Contour X=%d  Y=%d  W=%d  H=%d  Ratio=%f",
					rect.x, rect.y, rect.height, rect.width, ratio);
			Robot::robotLog->Write(robotLog);

			if (ratio < 0.3 || ratio > 0.5) continue;
			targets.push_back(contour);
		}

		targetCount = targets.size();

		if (targetCount == 2) {
			int maxHeight = 0;
			int maxX = 0;
			int minX = 320;

			for (std::vector<cv::Point> target: targets) {
				cv::Rect rect = boundingRect(target);
				if (minX > rect.x) minX = rect.x;
				if (maxX < (rect.x + rect.width)) maxX = rect.x + rect.width;
				if (maxHeight < rect.height) maxHeight = rect.height;
			}

			ratio = (double)(maxX - minX) / (double)maxHeight;

			if (ratio < 1.6 || ratio > 2.4) {
				targetCount = 0;
			} else {
				double centerX = ((double)(maxX - minX) / 2 + (double)minX - 160) / 160;
				targetAngle = (atan2(centerX, 2.11) * 180 / 3.14159);
				targetDistance = 3462 / (double)(maxX - minX);
			}
		}

	} else {
		Robot::robotLog->Write("Vision:  No Image from Camera");
	}

	host->SearchResults((targetCount == 2), targetAngle, targetDistance);

	imageA.deallocate();
	imageB.deallocate();
}
